#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include "cksum.h"
#include <sys/types.h>
#include "plfheader.h"
#include <stdio.h>
#include <stdlib.h>
#include "gzip.h"
#include <zlib.h>
#include "ini.h"
unsigned int readWord(int fd) {
	unsigned char b[4];
	read(fd, b, 4);
	return b[0]+(b[1]<<8)+(b[2]<<16)+(b[3]<<24);
}

void writeWord(int fd, unsigned int w) {
	unsigned char buff[4];
	buff[3]=(w>>24)&0xff;
	buff[2]=(w>>16)&0xff;
	buff[1]=(w>>8)&0xff;
	buff[0]=(w)&0xff;
	write(fd, buff, 4);
}

void readWords(int fd, unsigned int *where, int len) {
	int x;
	for (x=0; x<len; x++) where[x]=readWord(fd);
}

void writeWords(int fd, unsigned int *what, int len) {
	int x;
	for (x=0; x<len; x++) writeWord(fd, what[x]);
}

#define readPlfHeader(fd, header) readWords(fd, (unsigned int *)header, 14);
#define writePlfHeader(fd, header) writeWords(fd, (unsigned int *)header, 14);
#define readSectionHeader(fd, header) readWords(fd, (unsigned int *)header, 5);
#define writeSectionHeader(fd, header) writeWords(fd, (unsigned int *)header, 5);

int checkPlf(char *fileName) {
	int fd;
	int x;
	off_t pos1, pos2;
	int nocorr=0;
	plfFileHeader fHead;
	plfSectionHeader sHead;
	unsigned char *chunk;
	unsigned int crc;

	fd=open(fileName,O_RDWR);
	readPlfHeader(fd, &fHead);
	printf("Plf is ver %i.%i.%i.\n", fHead.verMaj, fHead.verMin, fHead.verRev);
	//Find out filesize & compare
	x=lseek(fd, 0, SEEK_END);
	if (x!=fHead.fileLen) {
		printf("Data len %i != len in header %i. Correcting.\n",x,fHead.fileLen);
		fHead.fileLen=x;
		nocorr++;
	}
	lseek(fd, 0, SEEK_SET);	
	writePlfHeader(fd, &fHead);

	while (lseek(fd, 0, SEEK_CUR)<fHead.fileLen) {
		printf("-- Chunk ---\n");
		pos1=lseek(fd, 0, SEEK_CUR);
		printf("Current pos in file: 0x%lX\n",(long) pos1);
		readSectionHeader(fd, &sHead);
		printf("Type=%i, len=%i, loadAddr=%X\n", sHead.type, sHead.length, sHead.loadAddr);
		chunk=malloc(sHead.length);
		read(fd, chunk, sHead.length);
		crc=cksum(chunk, sHead.length);
		int tst=open("test.out",O_WRONLY|O_CREAT|O_TRUNC,0644);
		write(tst,chunk,sHead.length);
		close(tst);
		free(chunk);
		if (crc!=sHead.checksum) {
			printf("Checksum doesn't match! calc=%X, file=%X. Correcting.\n",crc, sHead.checksum);
			sHead.checksum=crc;
			nocorr++;
		}
		//Write back header
		pos2=lseek(fd, 0, SEEK_CUR);
		lseek(fd, pos1, SEEK_SET);
		writeSectionHeader(fd, &sHead)
		lseek(fd, pos2, SEEK_SET);
		printf("Start=%x end=%x\n",(int)pos1,(int)pos2);

		//Seek to next 4-byte aligned pos
		int skip=lseek(fd, 0, SEEK_CUR)&3;
		printf("Skipping %i null bytes.\n",4-skip);
		if (skip!=0) lseek(fd, 4-skip, SEEK_CUR);
	}
	printf("All done; corrected %i things.\n",nocorr);
	close(fd);
	return nocorr;
}

void createPlf(char *filename, char *iniFile) {
	iniHandle ini;
	int fd,a;
	int fileno;
	char sect[100];
	int noFiles;
	plfFileHeader fHead;
	int addrPtr=0x40000;

	ini=iniOpenFile(iniFile);
	
	//First, create a quick&dirty plf. Checksums etc will be corrected Later(tm).
	fd=open(filename, O_WRONLY|O_CREAT|O_TRUNC,0644);
	fHead.magic=0x21464C50;
	fHead.unk1=8;
	fHead.unk2=0x38;
	fHead.unk3=0x14;
	fHead.unk4=0;
	fHead.unk5=0;
	fHead.type=iniGetInt(ini, "plf", "type");
	if (fHead.type==1) {
		fHead.entryPoint=iniGetInt(ini, "plf", "entrypoint");
	} else {
		fHead.entryPoint=0;
	}
	fHead.devType=iniGetInt(ini, "plf", "devtype");
	fHead.modelNo=iniGetInt(ini, "plf", "modelno");
	fHead.verMaj=iniGetInt(ini, "plf", "vermaj");
	fHead.verMin=iniGetInt(ini, "plf", "vermin");
	fHead.verRev=iniGetInt(ini, "plf", "verrev");
	noFiles=iniGetInt(ini, "plf", "filecount");
	writePlfHeader(fd, &fHead);

	for (fileno=1; fileno<=noFiles; fileno++) {
		plfSectionHeader sHead;
		int chunkLen;
		char *chunk;
		chunk=malloc(16*1024*1024); //overkill :)
		memset(&sHead, 0, sizeof(sHead));
		sprintf(sect,"file%i",fileno);
		sHead.type=iniGetInt(ini, sect, "type");
		if (sHead.type==0 || sHead.type==1 ) {
			//Code. Just add the file indicated.
			char fileName[255];
			int fin;
			int zeros=iniGetInt(ini, sect, "zerofill");
			printf("zeros: %i\n",zeros);
			if (zeros<=0) {
				iniGetStr(ini, sect, "file", fileName);
				fin=open(fileName, O_RDONLY);
				if (fin<0) {
					perror(fileName);
					exit(1);
				}
				chunkLen=read(fin, chunk, 16*1024*1024);
				close(fin);
			} else {
				memset(chunk, 0, zeros);
				chunkLen=zeros;
			}
			a=iniGetInt(ini, sect, "loadaddr");
			if (a>0) sHead.loadAddr=a; else sHead.loadAddr=addrPtr;
			addrPtr=sHead.loadAddr+chunkLen;
		} else if (sHead.type==2) {
			//No Idea. This is a 4-byte (one word) value containing a mysterious value.
			//Maybe the amount of ram needed?
			chunk[0]=0x58;
			chunk[1]=0xF2;
			chunk[2]=0x05;
			chunk[3]=0x00;
			chunkLen=4;
			a=iniGetInt(ini, sect, "loadaddr");
			if (a>0) sHead.loadAddr=a; else sHead.loadAddr=addrPtr;
			addrPtr=sHead.loadAddr+chunkLen;
		} else if (sHead.type==4) {
			//Gzipped/verbatim data. Don't bother with compression, just include the data verbatim.
			char fileName[255];
			int fin;
			iniGetStr(ini, sect, "path", chunk);
			iniGetStr(ini, sect, "file", fileName);
			fin=open(fileName, O_RDONLY);
			if (fin<0) {
				perror(fileName);
				exit(1);
			}
			chunkLen=strlen(chunk)+1;
			chunkLen+=read(fin, chunk+chunkLen, 16*1024*1024-chunkLen);
			close(fin);
		} else if (sHead.type==5) {
			iniGetStr(ini, sect, "path", chunk);
			chunkLen=strlen(chunk)+1;
		} else {
			printf("Don't know how to create a section of type %i!\n",sHead.type);
			exit(1);
		}
		sHead.length=chunkLen;
		writeSectionHeader(fd, &sHead);
		write(fd, chunk, chunkLen);
		//Pad up to next 4-byte aligned pos
		int skip=lseek(fd, 0, SEEK_CUR)&3;
		if (skip!=0) write(fd, "\000\000\000\000", 4-skip);
	}
	close(fd);
	checkPlf(filename);
}

void extractPlf(char *fileName) {
	int fd;
	unsigned long pos;
	char *chunk;
	plfFileHeader fHead;
	plfSectionHeader sHead;

	fd=open(fileName,O_RDONLY);
	readPlfHeader(fd, &fHead);

	while (lseek(fd, 0, SEEK_CUR)<fHead.fileLen) {
		pos=lseek(fd, 0, SEEK_CUR);
		readSectionHeader(fd, &sHead);
		chunk=malloc(sHead.length);
		read(fd, chunk, sHead.length);

		if (sHead.type==0) {
			//Code.
		} else if (sHead.type==4) {
			if (sHead.uncompSize==0) {
				//Not compressed. Just write out.
				printf("Writing verbatim data to %s\n", chunk);
				int fout=open(chunk, O_WRONLY|O_CREAT|O_TRUNC, 0644);
				int skip=strlen(chunk)+1;
				write(fout, chunk+skip, sHead.length-skip);
				close(fout);
			} else {
				char *deflated;
				unsigned long len;
				len=sHead.uncompSize*2+20; //gunzip may need the overhead
				deflated=malloc(len);
				int r=gzDecompress(deflated, &len, chunk, sHead.length);
				if (r==Z_BUF_ERROR) {
					printf("Error ungzipping chunk: data is larger than specified in header!\n");
				} else if (r==Z_DATA_ERROR) {
					printf("Error ungzipping chunk: data is corrupted!\n");
					printf("Offending chunk is at offset %li.\n",pos);
				} else if (r==Z_OK) {
					if (sHead.uncompSize!=len) {
						printf("Warning: chunk len is %lu, but header tells me it's %u.\n", len, sHead.uncompSize);
					}
					printf("Writing uncompressed data to %s\n", deflated);
					int fout=open(deflated, O_WRONLY|O_CREAT|O_TRUNC, 0644);
					int skip=strlen(deflated)+1;
					write(fout, deflated+skip, len-skip);
					close(fout);
					free(deflated);
				} else {
					printf("Error ungzipping chunk: unknown error %i!\n", r);
				}
			}
		} else if (sHead.type==5) {
			printf("Creating dir %s\n",chunk);
			mkdir(chunk,0755);
		} else {
			printf("Can't parse chunk type %i. Are you sure this is a plf?\n", sHead.type);
		}
		free(chunk);
		//Seek to next 4-byte aligned pos
		int skip=lseek(fd, 0, SEEK_CUR)&3;
		if (skip!=0) lseek(fd, 4-skip, SEEK_CUR);
	}
}

int main(int argc, char** argv) {
	if (argc<3) {
		printf("Usage: %s [-c|-x|-t] file.plf [inifile.ini]", argv[0]);
		printf(" -c: create plf from inifile\n");
		printf(" -x: extract plf\n");
		printf(" -t: test plf\n");
		exit(0);
	}
	if (strcmp(argv[1],"-x")==0) extractPlf(argv[2]);
	if (strcmp(argv[1],"-t")==0) checkPlf(argv[2]);
	if (strcmp(argv[1],"-c")==0) createPlf(argv[2], argv[3]);
	exit(0);
}